import math

import torch
import torch.nn as nn


def gelu(x):
    return 0.5 * x * (1 + torch.tanh(math.sqrt(2 / math.pi) * (x + 0.044715 * torch.pow(x, 3))))


class PositionwiseFeedForward(nn.Module):
    """ A two-layer Feed-Forward-Network with residual layer norm.

    Args:
        d_model (int): the size of input for the first-layer of the FFN.
        d_ff (int): the hidden layer size of the second-layer
            of the FNN.
        dropout (float): dropout probability in :math:`[0, 1)`.
    """

    def __init__(self, d_model, d_ff, dropout=0.1):
        super(PositionwiseFeedForward, self).__init__()
        self.w_1 = nn.Linear(d_model, d_ff)
        self.w_2 = nn.Linear(d_ff, d_model)
        self.layer_norm = nn.LayerNorm(d_model, eps=1e-6)
        self.actv = gelu
        self.dropout_1 = nn.Dropout(dropout)
        self.dropout_2 = nn.Dropout(dropout)

    def forward(self, x):
        inter = self.dropout_1(self.actv(self.w_1(self.layer_norm(x))))
        output = self.dropout_2(self.w_2(inter))
        return output + x


class MultiHeadedAttention(nn.Module):
    """
    Multi-Head Attention module from
    "Attention is All You Need"
    :cite:`DBLP:journals/corr/VaswaniSPUJGKP17`.

    Similar to standard `dot` attention but uses
    multiple attention distributions simulataneously
    to select relevant items.

    .. mermaid::

       graph BT
          A[key]
          B[value]
          C[query]
          O[output]
          subgraph Attn
            D[Attn 1]
            E[Attn 2]
            F[Attn N]
          end
          A --> D
          C --> D
          A --> E
          C --> E
          A --> F
          C --> F
          D --> O
          E --> O
          F --> O
          B --> O

    Also includes several additional tricks.

    Args:
       head_count (int): number of parallel heads
       model_dim (int): the dimension of keys/values/queries,
           must be divisible by head_count
       dropout (float): dropout parameter
    """

    def __init__(self, head_count, model_dim, dropout=0.1, use_final_linear=True):
        assert model_dim % head_count == 0
        self.dim_per_head = model_dim // head_count
        self.model_dim = model_dim

        super(MultiHeadedAttention, self).__init__()
        self.head_count = head_count

        self.linear_keys = nn.Linear(model_dim,
                                     head_count * self.dim_per_head)
        self.linear_values = nn.Linear(model_dim,
                                       head_count * self.dim_per_head)
        self.linear_query = nn.Linear(model_dim,
                                      head_count * self.dim_per_head)
        self.softmax = nn.Softmax(dim=-1)
        self.dropout = nn.Dropout(dropout)
        self.use_final_linear = use_final_linear
        if (self.use_final_linear):
            self.final_linear = nn.Linear(model_dim, model_dim)

    def forward(self, key, value, query, mask=None,
                layer_cache=None, type=None, predefined_graph_1=None):
        """
        Compute the context vector and the attention vectors.

        Args:
           key (`FloatTensor`): set of `key_len`
                key vectors `[batch, key_len, dim]`
           value (`FloatTensor`): set of `key_len`
                value vectors `[batch, key_len, dim]`
           query (`FloatTensor`): set of `query_len`
                 query vectors  `[batch, query_len, dim]`
           mask: binary mask indicating which keys have
                 non-zero attention `[batch, query_len, key_len]`
        Returns:
           (`FloatTensor`, `FloatTensor`) :

           * output context vectors `[batch, query_len, dim]`
           * one of the attention vectors `[batch, query_len, key_len]`
        """

        # CHECKS
        # batch, k_len, d = key.size()
        # batch_, k_len_, d_ = value.size()
        # aeq(batch, batch_)
        # aeq(k_len, k_len_)
        # aeq(d, d_)
        # batch_, q_len, d_ = query.size()
        # aeq(batch, batch_)
        # aeq(d, d_)
        # aeq(self.model_dim % 8, 0)
        # if mask is not None:
        #    batch_, q_len_, k_len_ = mask.size()
        #    aeq(batch_, batch)
        #    aeq(k_len_, k_len)
        #    aeq(q_len_ == q_len)
        # END CHECKS

        batch_size = key.size(0)
        dim_per_head = self.dim_per_head
        head_count = self.head_count
        key_len = key.size(1)
        query_len = query.size(1)

        def shape(x):
            """  projection """
            return x.view(batch_size, -1, head_count, dim_per_head) \
                .transpose(1, 2)

        def unshape(x):
            """  compute context """
            return x.transpose(1, 2).contiguous() \
                .view(batch_size, -1, head_count * dim_per_head)

        # 1) Project key, value, and query.
        if layer_cache is not None:
            if type == "self":
                query, key, value = self.linear_query(query), \
                                    self.linear_keys(query), \
                                    self.linear_values(query)

                key = shape(key)
                value = shape(value)

                if layer_cache is not None:
                    device = key.device
                    if layer_cache["self_keys"] is not None:
                        key = torch.cat(
                            (layer_cache["self_keys"].to(device), key),
                            dim=2)
                    if layer_cache["self_values"] is not None:
                        value = torch.cat(
                            (layer_cache["self_values"].to(device), value),
                            dim=2)
                    layer_cache["self_keys"] = key
                    layer_cache["self_values"] = value
            elif type == "context":
                query = self.linear_query(query)
                if layer_cache is not None:
                    if layer_cache["memory_keys"] is None:
                        key, value = self.linear_keys(key), \
                                     self.linear_values(value)
                        key = shape(key)
                        value = shape(value)
                    else:
                        key, value = layer_cache["memory_keys"], \
                                     layer_cache["memory_values"]
                    layer_cache["memory_keys"] = key
                    layer_cache["memory_values"] = value
                else:
                    key, value = self.linear_keys(key), \
                                 self.linear_values(value)
                    key = shape(key)
                    value = shape(value)
        else:
            key = self.linear_keys(key)
            value = self.linear_values(value)
            query = self.linear_query(query)
            key = shape(key)
            value = shape(value)

        query = shape(query)

        key_len = key.size(2)
        query_len = query.size(2)

        # 2) Calculate and scale scores.
        query = query / math.sqrt(dim_per_head)
        scores = torch.matmul(query, key.transpose(2, 3))

        if mask is not None:
            mask = mask.unsqueeze(1).expand_as(scores)
            scores = scores.masked_fill(mask, -1e18)

        # 3) Apply attention dropout and compute context vectors.

        attn = self.softmax(scores)

        if (not predefined_graph_1 is None):
            attn_masked = attn[:, -1] * predefined_graph_1
            attn_masked = attn_masked / (torch.sum(attn_masked, 2).unsqueeze(2) + 1e-9)

            attn = torch.cat([attn[:, :-1], attn_masked.unsqueeze(1)], 1)

        drop_attn = self.dropout(attn)
        if (self.use_final_linear):
            context = unshape(torch.matmul(drop_attn, value))
            output = self.final_linear(context)
            return output
        else:
            context = torch.matmul(drop_attn, value)
            return context

        # CHECK
        # batch_, q_len_, d_ = output.size()
        # aeq(q_len, q_len_)
        # aeq(batch, batch_)
        # aeq(d, d_)

        # Return one attn




import math

import torch
import torch.nn as nn

from .neural import MultiHeadedAttention, PositionwiseFeedForward
from .rnn import LayerNormLSTM


class Classifier(nn.Module):
    def __init__(self, hidden_size):
        super(Classifier, self).__init__()
        self.linear1 = nn.Linear(hidden_size, 1)
        self.sigmoid = nn.Sigmoid()

    def forward(self, x, mask_cls):
        h = self.linear1(x).squeeze(-1)
        sent_scores = self.sigmoid(h) * mask_cls.float()
        return sent_scores


class PositionalEncoding(nn.Module):

    def __init__(self, dropout, dim, max_len=5000):
        pe = torch.zeros(max_len, dim)
        position = torch.arange(0, max_len).unsqueeze(1)
        div_term = torch.exp((torch.arange(0, dim, 2, dtype=torch.float) *
                              -(math.log(10000.0) / dim)))
        pe[:, 0::2] = torch.sin(position.float() * div_term)
        pe[:, 1::2] = torch.cos(position.float() * div_term)
        pe = pe.unsqueeze(0)
        super(PositionalEncoding, self).__init__()
        self.register_buffer('pe', pe)
        self.dropout = nn.Dropout(p=dropout)
        self.dim = dim

    def forward(self, emb, step=None):
        emb = emb * math.sqrt(self.dim)
        if (step):
            emb = emb + self.pe[:, step][:, None, :]

        else:
            emb = emb + self.pe[:, :emb.size(1)]
        emb = self.dropout(emb)
        return emb

    def get_emb(self, emb):
        return self.pe[:, :emb.size(1)]


class TransformerEncoderLayer(nn.Module):
    def __init__(self, d_model, heads, d_ff, dropout):
        super(TransformerEncoderLayer, self).__init__()

        self.self_attn = MultiHeadedAttention(
            heads, d_model, dropout=dropout)
        self.feed_forward = PositionwiseFeedForward(d_model, d_ff, dropout)
        self.layer_norm = nn.LayerNorm(d_model, eps=1e-6)
        self.dropout = nn.Dropout(dropout)

    def forward(self, iter, query, inputs, mask):
        if (iter != 0):
            input_norm = self.layer_norm(inputs)
        else:
            input_norm = inputs

        mask = mask.unsqueeze(1)
        context = self.self_attn(input_norm, input_norm, input_norm,
                                 mask=mask)
        out = self.dropout(context) + inputs
        return self.feed_forward(out)


class TransformerInterEncoder(nn.Module):
    def __init__(self, d_model, d_ff, heads, dropout, num_inter_layers=0):
        super(TransformerInterEncoder, self).__init__()
        self.d_model = d_model
        self.num_inter_layers = num_inter_layers
        self.pos_emb = PositionalEncoding(dropout, d_model)
        self.transformer_inter = nn.ModuleList(
            [TransformerEncoderLayer(d_model, heads, d_ff, dropout)
             for _ in range(num_inter_layers)])
        self.dropout = nn.Dropout(dropout)
        self.layer_norm = nn.LayerNorm(d_model, eps=1e-6)
        self.wo = nn.Linear(d_model, 1, bias=True)
        self.sigmoid = nn.Sigmoid()

    def forward(self, top_vecs, mask):
        """ See :obj:`EncoderBase.forward()`"""

        batch_size, n_sents = top_vecs.size(0), top_vecs.size(1)
        pos_emb = self.pos_emb.pe[:, :n_sents]
        x = top_vecs * mask[:, :, None].float()
        x = x + pos_emb

        for i in range(self.num_inter_layers):
            #x = self.transformer_inter[i](i, x, x, 1 - mask)  # all_sents * max_tokens * dim
            x = self.transformer_inter[i](i, x, x, ~mask)

        x = self.layer_norm(x)
        sent_scores = self.sigmoid(self.wo(x))
        sent_scores = sent_scores.squeeze(-1) * mask.float()

        return sent_scores


class RNNEncoder(nn.Module):

    def __init__(self, bidirectional, num_layers, input_size,
                 hidden_size, dropout=0.0):
        super(RNNEncoder, self).__init__()
        num_directions = 2 if bidirectional else 1
        assert hidden_size % num_directions == 0
        hidden_size = hidden_size // num_directions

        self.rnn = LayerNormLSTM(
            input_size=input_size,
            hidden_size=hidden_size,
            num_layers=num_layers,
            bidirectional=bidirectional)
        self.wo = nn.Linear(num_directions * hidden_size, 1, bias=True)
        self.dropout = nn.Dropout(dropout)
        self.sigmoid = nn.Sigmoid()

    def forward(self, x, mask):
        """See :func:`EncoderBase.forward()`"""
        x = torch.transpose(x, 1, 0)
        memory_bank, _ = self.rnn(x)
        memory_bank = self.dropout(memory_bank) + x
        memory_bank = torch.transpose(memory_bank, 1, 0)

        sent_scores = self.sigmoid(self.wo(memory_bank))
        sent_scores = sent_scores.squeeze(-1) * mask.float()
        return sent_scores




import torch
import torch.nn as nn
from transformers import XLMRobertaModel,XLMRobertaConfig
from torch.nn.init import xavier_uniform_

from .encoder import TransformerInterEncoder, Classifier, RNNEncoder
from .optimizers import Optimizer


def build_optim(args, model, checkpoint):
    """ Build optimizer """
    saved_optimizer_state_dict = None

    if args.train_from != '':
        optim = checkpoint['optim']
        saved_optimizer_state_dict = optim.optimizer.state_dict()
    else:
        optim = Optimizer(
            args.optim, args.lr, args.max_grad_norm,
            beta1=args.beta1, beta2=args.beta2,
            decay_method=args.decay_method,
            warmup_steps=args.warmup_steps)

    optim.set_parameters(list(model.named_parameters()))

    if args.train_from != '':
        optim.optimizer.load_state_dict(saved_optimizer_state_dict)
        if args.visible_gpus != '-1':
            for state in optim.optimizer.state.values():
                for k, v in state.items():
                    if torch.is_tensor(v):
                        state[k] = v.cuda()

        if (optim.method == 'adam') and (len(optim.optimizer.state) < 1):
            raise RuntimeError(
                "Error: loaded Adam optimizer from existing model" +
                " but optimizer state is empty")

    return optim


class XLM_R(nn.Module):
    def __init__(self, temp_dir, load_pretrained_bert, xlmr_config):
        super(XLM_R, self).__init__()
        if(load_pretrained_bert):
            self.model = XLMRobertaModel.from_pretrained("hfl/cino-large-v2")
        else:
            self.model = XLMRobertaModel(xlmr_config)

    def forward(self, x, segs, mask):

        outputs = self.model(input_ids = x,attention_mask = mask)
        last_hidden_state = outputs.last_hidden_state
        
        top_vec = last_hidden_state
        return top_vec



class Summarizer(nn.Module):
    def __init__(self, args, device, load_pretrained_bert = False, xlmr_config = None):
        super(Summarizer, self).__init__()
        self.args = args
        self.device = device
        self.xlmr = XLM_R(args.temp_dir, load_pretrained_bert, xlmr_config)
        if (args.encoder == 'classifier'):
            self.encoder = Classifier(self.xlmr.model.config.hidden_size)
        elif(args.encoder=='transformer'):
            self.encoder = TransformerInterEncoder(self.xlmr.model.config.hidden_size, args.ff_size, args.heads,
                                                   args.dropout, args.inter_layers)
        elif(args.encoder=='rnn'):
            self.encoder = RNNEncoder(bidirectional=True, num_layers=1,
                                      input_size=self.xlmr.model.config.hidden_size, hidden_size=args.rnn_size,
                                      dropout=args.dropout)
        elif (args.encoder == 'baseline'):
            xlmr_config = XLMRobertaConfig.from_pretrained("hfl/cino-large-v2")
            self.xlmr.model = XLMRobertaModel(xlmr_config)
            self.encoder = Classifier(self.xlmr.model.config.hidden_size)

        if args.param_init != 0.0:
            for p in self.encoder.parameters():
                p.data.uniform_(-args.param_init, args.param_init)
        if args.param_init_glorot:
            for p in self.encoder.parameters():
                if p.dim() > 1:
                    xavier_uniform_(p)
        self.to(device)

    def load_cp(self, pt):
        self.load_state_dict(pt['model'], strict=True)

    def forward(self, x, segs, clss, mask, mask_cls, sentence_range=None):
        
        top_vec = self.xlmr(x, segs, mask)

        sents_vec = top_vec[torch.arange(top_vec.size(0)).unsqueeze(1), clss]
        sents_vec = sents_vec * mask_cls[:, :, None].float()
        sent_scores = self.encoder(sents_vec, mask_cls).squeeze(-1)
        return sent_scores, mask_cls
    



import torch
import torch.nn.functional as F
from torch import nn


class LayerNormLSTMCell(nn.LSTMCell):

    def __init__(self, input_size, hidden_size, bias=True):
        super().__init__(input_size, hidden_size, bias)

        self.ln_ih = nn.LayerNorm(4 * hidden_size)
        self.ln_hh = nn.LayerNorm(4 * hidden_size)
        self.ln_ho = nn.LayerNorm(hidden_size)

    def forward(self, input, hidden=None):
        
        if hidden is None:
            hx = input.new_zeros(input.size(0), self.hidden_size, requires_grad=False)
            cx = input.new_zeros(input.size(0), self.hidden_size, requires_grad=False)
        else:
            hx, cx = hidden
    

        gates = self.ln_ih(F.linear(input, self.weight_ih, self.bias_ih)) \
                + self.ln_hh(F.linear(hx, self.weight_hh, self.bias_hh))
        i, f, o = gates[:, :(3 * self.hidden_size)].sigmoid().chunk(3, 1)
        g = gates[:, (3 * self.hidden_size):].tanh()

        cy = (f * cx) + (i * g)
        hy = o * self.ln_ho(cy).tanh()
        return hy, cy


class LayerNormLSTM(nn.Module):

    def __init__(self, input_size, hidden_size, num_layers=1, bias=True, bidirectional=False):
        super().__init__()
        self.input_size = input_size
        self.hidden_size = hidden_size
        self.num_layers = num_layers
        self.bidirectional = bidirectional

        num_directions = 2 if bidirectional else 1
        self.hidden0 = nn.ModuleList([
            LayerNormLSTMCell(input_size=(input_size if layer == 0 else hidden_size * num_directions),
                              hidden_size=hidden_size, bias=bias)
            for layer in range(num_layers)
        ])

        if self.bidirectional:
            self.hidden1 = nn.ModuleList([
                LayerNormLSTMCell(input_size=(input_size if layer == 0 else hidden_size * num_directions),
                                  hidden_size=hidden_size, bias=bias)
                for layer in range(num_layers)
            ])

    def forward(self, input, hidden=None):
        seq_len, batch_size, hidden_size = input.size()  # supports TxNxH only
        num_directions = 2 if self.bidirectional else 1
        if hidden is None:
            hx = input.new_zeros(self.num_layers * num_directions, batch_size, self.hidden_size, requires_grad=False)
            cx = input.new_zeros(self.num_layers * num_directions, batch_size, self.hidden_size, requires_grad=False)
        else:
            hx, cx = hidden

        ht = [[None, ] * (self.num_layers * num_directions)] * seq_len
        ct = [[None, ] * (self.num_layers * num_directions)] * seq_len

        if self.bidirectional:
            xs = input
            for l, (layer0, layer1) in enumerate(zip(self.hidden0, self.hidden1)):
                l0, l1 = 2 * l, 2 * l + 1
                h0, c0, h1, c1 = hx[l0], cx[l0], hx[l1], cx[l1]
                for t, (x0, x1) in enumerate(zip(xs, reversed(xs))):
                    ht[t][l0], ct[t][l0] = layer0(x0, (h0, c0))
                    h0, c0 = ht[t][l0], ct[t][l0]
                    t = seq_len - 1 - t
                    ht[t][l1], ct[t][l1] = layer1(x1, (h1, c1))
                    h1, c1 = ht[t][l1], ct[t][l1]
                xs = [torch.cat((h[l0], h[l1]), dim=1) for h in ht]
            y = torch.stack(xs)
            hy = torch.stack(ht[-1])
            cy = torch.stack(ct[-1])
        else:
            h, c = hx, cx
            for t, x in enumerate(input):
                for l, layer in enumerate(self.hidden0):
                    ht[t][l], ct[t][l] = layer(x, (h[l], c[l]))
                    x = ht[t][l]
                h, c = ht[t], ct[t]
            y = torch.stack([h[-1] for h in ht])
            hy = torch.stack(ht[-1])
            cy = torch.stack(ct[-1])

        return y, (hy, cy)




import torch
import copy
import numpy as np
import re
from transformers import XLMRobertaModel,XLMRobertaConfig,XLMRobertaTokenizer
from types import SimpleNamespace
from torch.utils.data import DataLoader, TensorDataset
from torch.utils.data import Dataset
from tqdm import tqdm

tokenizer = XLMRobertaTokenizer.from_pretrained("hfl/cino-large-v2")
def truncate_context_to_fit_tokenizer(context, tokenizer, max_length=512):
    # 首先对原始的context进行分词
    context_tokens = tokenizer.tokenize(context)
    # 如果分词后的长度小于或等于max_length，就直接返回原始context
    if len(context_tokens) < max_length:
        return context

    # 如果分词后的长度大于max_length，需要截断
    # 注意：这里我们不再添加空格，而是直接截取token列表
    truncated_tokens = context_tokens[:max_length]

    # 使用分词器的convert_tokens_to_string方法将token列表转换回字符串
    truncated_context = tokenizer.convert_tokens_to_string(truncated_tokens)

    #print(truncated_context)

    return truncated_context

def normal_cut_sentence(text):
    text = re.sub('([。！？\?])([^’”])',r'\1\n\2',text)#普通断句符号且后面没有引号
    text = re.sub('(\.{6})([^’”])',r'\1\n\2',text)#英文省略号且后面没有引号
    text = re.sub('(\…{2})([^’”])',r'\1\n\2',text)#中文省略号且后面没有引号
    text = re.sub('([.。！？\?\.{6}\…{2}][’”])([^’”])',r'\1\n\2',text)#断句号+引号且后面没有引号
    return text.split("\n")

def tokenize_zh_better(text):
    p = re.compile("“.*?”")
    list = []
    index = 0
    length = len(text)
    for i in p.finditer(text):
        temp = ''
        start = i.start()
        end = i.end()
        for j in range(index, start):
            temp += text[j]
        if temp != '':
            temp_list = normal_cut_sentence(temp)
            list += temp_list
        temp = ''
        for k in range(start, end):
            temp += text[k]
        if temp != ' ':
            list.append(temp)
        index = end
    return list

def tokenize_zh_base(para):
    para = re.sub('([。！？\?])([^”’])', r"\1\n\2", para)  # 单字符断句符
    para = re.sub('(\.{6})([^”’])', r"\1\n\2", para)  # 英文省略号
    para = re.sub('(\…{2})([^”’])', r"\1\n\2", para)  # 中文省略号
    para = re.sub('([。！？\?][”’])([^，。！？\?])', r'\1\n\2', para)
    # 如果双引号前有终止符，那么双引号才是句子的终点，把分句符\n放到双引号后，注意前面的几句都小心保留了双引号
    para = para.rstrip()  # 段尾如果有多余的\n就去掉它
    # 很多规则中会考虑分号;，但是这里我把它忽略不计，破折号、英文双引号等同样忽略，需要的再做些简单调整即可。
    return para.split("\n")


def tokenize_bo(text):
    # 这里假设藏文句子结束可以用特殊的藏文标点符号 "།" 来识别
    sentences = text.split('།')
    # 去除句子两边的空白并过滤掉空句子
    sentences = [sentence.strip() for sentence in sentences if sentence.strip()]
    return sentences

def tokenize_ug(text):
    # 这里假设维文句子结束可以用句号 "。"、问号 "؟" 和感叹号 "！" 来识别
    # 注意维吾尔文中使用的是阿拉伯问号 "؟" 而不是标准问号 "?"
    sentences = []
    sentence = ''
    for char in text:
        sentence += char
        if char in ('。', '؟', '！','.'):
            sentences.append(sentence.strip())
            sentence = ''
    # 确保最后一个句子被添加，即使它后面没有标点符号
    if sentence:
        sentences.append(sentence.strip())
    return sentences


def sentence_tokenize(text, language='zh'):
    """
    Tokenize a text into sentences for the specified language.

    Parameters:
    text (str): The text to be tokenized.
    language (str): The language code for the tokenizer. Supported languages are 'zh', 'bo', and 'ug'.

    Returns:
    list: A list of sentences.
    """
    
    # 首先检查 text 是否为字符串
    if not isinstance(text, str):
        raise ValueError("Input must be a string.")
    
    

    # 根据传入的语言参数调用相应的分句方法
    if language == 'zh' or language == '中文':
        return tokenize_zh_base(text)
    elif language == 'bo' or language == '藏文':
        return tokenize_bo(text)
    elif language == 'ug' or language == '维文':
        return tokenize_ug(text)
    else:
        # 如果传入了一个未知的语言代码，抛出一个异常
        raise ValueError(f"Unsupported language code: {language}")



class CustomDataset(Dataset):
    def __init__(self, input_ids, segs, clss, attention_masks, mask_cls, src_str):
        self.input_ids = input_ids
        self.segs = segs
        self.clss = clss
        self.attention_masks = attention_masks
        self.mask_cls = mask_cls
        self.src_str = src_str

    def __len__(self):
        return len(self.input_ids)

    def __getitem__(self, index):
        return {
            'input_ids': self.input_ids[index],
            'segs': self.segs[index],
            'clss': self.clss[index],
            'attention_masks': self.attention_masks[index],
            'mask_cls': self.mask_cls[index],
            'src_str': self.src_str[index]  # 这里把句子也包括进来
        }

def custom_collate_fn(batch):
    # 假设batch是一个列表，其中每个元素是一个字典，拥有相同的键
    # 例如：[{'input_ids': tensor1, 'segs': tensor2, ..., 'src_str': list1}, {...}, ...]

    # 初始化字典以收集所有数据字段
    batch_data = {
        'input_ids': [],
        'segs': [],
        'clss': [],
        'attention_masks': [],
        'mask_cls': [],
        'src_str': []
    }

    for item in batch:
        batch_data['input_ids'].append(item['input_ids'])
        batch_data['segs'].append(item['segs'])
        batch_data['clss'].append(item['clss'])
        batch_data['attention_masks'].append(item['attention_masks'])
        batch_data['mask_cls'].append(item['mask_cls'])
        # src_str 是一个句子列表，直接加入列表
        batch_data['src_str'].append(item['src_str'])

    # 对于前五个字段，我们使用torch.stack来合并列表中的tensor
    for key in ['input_ids', 'segs', 'clss', 'attention_masks', 'mask_cls']:
        batch_data[key] = torch.stack(batch_data[key], dim=0)

    # src_str 保持为列表形式，不需要stack
    # batch_data['src_str'] 本来就已经是列表了，不需要额外处理

    return batch_data

def process_text_2_test_iter(contexts, device, len_pred = 2, language='zh', batch_size=1, print_setences=False):
    # 这将存储所有文本的处理结果
    all_input_ids = []
    all_attention_masks = []
    all_segs = []
    all_clss = []
    all_mask_cls = []
    all_src_str = []
    max_length = 512

    for context in contexts:

        #尽量保留原文本
        context = truncate_context_to_fit_tokenizer(context, tokenizer, max_length = max_length)
        
        # 分句
        sentences = sentence_tokenize(context, language)
        sum_sentences_length = sum((len(sentence) for sentence in sentences))

        if sum_sentences_length>=max_length:
            context = context[:max_length-1]
            sentences = sentence_tokenize(context, language)

        #assert len(sentences)<len_pred

        # tokenizer处理，获取input_ids和attention_mask
        encoding = tokenizer(context, max_length=max_length, padding='max_length', truncation=True, return_tensors='pt')
        input_ids = encoding['input_ids']
        attention_mask = encoding['attention_mask']
        
        # segs全零张量,CINO没有token_typeid
        segs = torch.zeros(max_length, dtype=torch.long)
        
        # clss为每个句子的起始位置
        clss = [-1] * input_ids.size(1)  # 初始化clss列表
        clss[0] = 0
        for i, sentence in enumerate(sentences):
            clss[i+1] = clss[i]+len(sentence)

            if print_setences:
                print(f'第:{i+1}个句子:{sentence}')
            
        
        clss = torch.tensor(clss)
        # mask_cls
        mask_cls = ~ (clss == -1)
        clss[clss == -1] = 0
        
        # 累积处理结果
        all_input_ids.append(input_ids)
        all_attention_masks.append(attention_mask)
        all_segs.append(segs)
        all_clss.append(clss)
        all_mask_cls.append(mask_cls)
        all_src_str.append(sentences)

    num_samples = len(all_input_ids)
    
    # Flatten all_input_ids, all_attention_masks, and all_segs, assuming they are lists of 2D tensors
    all_input_ids = torch.cat(all_input_ids, dim=0)
    all_attention_masks = torch.cat(all_attention_masks, dim=0)
    
    # Flatten all_clss and all_mask_cls, and make sure they match the batch dimension
    # Assuming all_clss and all_mask_cls are lists of 1D tensors
    all_segs = torch.cat([segs.unsqueeze(0) for segs in all_segs], dim=0)
    all_clss = torch.cat([cls.unsqueeze(0) for cls in all_clss], dim=0)
    all_mask_cls = torch.cat([mask.unsqueeze(0) for mask in all_mask_cls], dim=0)


    # Check that the batch dimension matches
    assert all_input_ids.size(0) == num_samples
    assert all_attention_masks.size(0) == num_samples
    assert all_segs.size(0) == num_samples
    assert all_clss.size(0) == num_samples
    assert all_mask_cls.size(0) == num_samples

    #print(f'all_input_ids:{all_input_ids.shape}')
    #print(all_src_str)
    #print(len(all_src_str))
    # 创建TensorDatasets
    dataset = CustomDataset(
        all_input_ids,
        all_segs,
        all_clss,
        all_attention_masks,
        all_mask_cls,
        all_src_str
    )
    
    # 使用DataLoader来处理batching
    data_loader = DataLoader(dataset, batch_size=batch_size, collate_fn=custom_collate_fn, shuffle=False)

    
    return data_loader



def generate(model, contexts, device, batch_size = 1, len_pred = 2, language = 'zh'):
        """ Validate model.
            valid_iter: validate data iterator
        Returns:
            :obj:`nmt.Statistics`: validation loss statistics
        """
        # Set model in validating mode.
        def _get_ngrams(n, text):
            ngram_set = set()
            text_length = len(text)
            max_index_ngram_start = text_length - n
            for i in range(max_index_ngram_start + 1):
                ngram_set.add(tuple(text[i:i + n]))
            return ngram_set

        def _block_tri(c, p):
            tri_c = _get_ngrams(3, c.split())
            for s in p:
                tri_s = _get_ngrams(3, s.split())
                if len(tri_c.intersection(tri_s))>0:
                    return True
            return False
        
        pred_list = []
        
        test_iter = process_text_2_test_iter(contexts, device, batch_size=batch_size, len_pred=len_pred, language = language)
        model.eval()
        with torch.no_grad():
            
            for batch in tqdm(test_iter, desc="Processing", leave=True):
                
                src = batch['input_ids'].to(device)
                segs = batch['segs'].to(device)
                clss = batch['clss'].to(device)
                mask = batch['attention_masks'].to(device)
                mask_cls = batch['mask_cls'].to(device)
                src_str = batch['src_str']

                # src = batch[0].to(device)
                # segs = batch[1].to(device)
                # clss = batch[2].to(device)
                # mask = batch[3].to(device)
                # mask_cls = batch[4].to(device)
                

                
                pred = []
                              
                sent_scores, mask = model(src, segs, clss, mask, mask_cls)                           
                sent_scores = sent_scores + mask.float()
                sent_scores = sent_scores.cpu().data.numpy()
                selected_ids = np.argsort(-sent_scores, 1)

                for i, idx in enumerate(selected_ids):
                    _pred = []
                    if(len(src_str[i])==0):
                        continue
                    
                    for j in selected_ids[i][:len(src_str[i])]:
                        if(j>=len(src_str[i])):
                            continue

                        candidate = src_str[i][j].strip()
                        if(not _block_tri(candidate,_pred)):
                                _pred.append(candidate)
                        

                        if (len(_pred) == len_pred):
                            break

                    pred_list.append(copy.copy(_pred))

                    _pred = '<q>'.join(_pred)

                    pred.append(_pred)
                
        return pred_list

            


from .model_builder import *
from .generate import *
import torch
from types import SimpleNamespace

args_dict = {
    "encoder": 'transformer',
    "mode": 'test',
    "bert_data_path": '/',
    "model_path": '../',
    "result_path": '../',
    "temp_dir": '/',
    "bert_config_path": '/',
    "batch_size": 1000,
    "use_interval": True,
    "hidden_size": 2048,
    "ff_size": 512,
    "heads": 4,
    "inter_layers": 2,
    "rnn_size": 1024,
    "param_init": 0,
    "param_init_glorot": True,
    "dropout": 0.1,
    "optim": 'adam',
    "lr": 1,
    "beta1": 0.9,
    "beta2": 0.999,
    "decay_method": '',
    "warmup_steps": 10000,
    "max_grad_norm": 0,
    "save_checkpoint_steps": 5,
    "accum_count": 1,
    "world_size": 1,
    "report_every": 1,
    "train_steps": 1000,
    "recall_eval": False,
    "visible_gpus": '-1',
    "gpu_ranks": '0',
    "log_file": '/',
    "dataset": '',
    "seed": 3407,
    "test_all": False,
    "test_from": '',
    "train_from": '',
    "report_rouge": True,
    "block_trigram": True
}



def initialize_args():
    args = SimpleNamespace(**args_dict)
    return args

class CINOSUM(nn.Module):
    """
    CINOSUM类，继承自PyTorch的nn.Module，用于多语言摘要生成任务。

    参数:
    - device (str): 指定模型运行的设备，默认为'cpu'。
    - model_path (str): 指定预训练模型的路径，默认为None值。
    - len_pred (int): 生成摘要的预期长度，默认为2。
    - language (str): 模型处理的目标语言，默认为'zh'（中文）。
    - load_pretrained_bert (bool): 是否加载训练的模型，默认为True。

    属性:
    - args: 通过initialize_args函数初始化的参数对象。
    - len_pred: 生成摘要的长度。
    - language: 模型处理的语言。
    - model_path: 预训练模型的文件路径。
    - config: 使用XLMRobertaConfig加载的模型配置对象。
    - model: Summarizer模型实例。
    - device: 模型运行的设备。
    - if_load: 标记模型是否已加载，避免重复加载。

    方法:
    - __init__: 类的构造函数，用于初始化模型及其相关配置。

    注意:
    - 若model_path为空或None，则会跳过加载模型的步骤，使用随机初始化的模型。
    - 需要事先导入PyTorch相关模块和XLMRobertaConfig，以及定义Summarizer模型。
    """
    def __init__(self, device='cpu', model_path = None, len_pred = 2, language = 'zh', load_pretrained_bert = True):
        super(CINOSUM, self).__init__()

        self.args = initialize_args()

        self.len_pred = len_pred
        self.language = language
        self.model_path = model_path

        self.config = XLMRobertaConfig.from_pretrained("hfl/cino-large-v2")
        self.model = Summarizer(self.args, device, xlmr_config=self.config)
        self.device = device
        
        self.if_load = False
        #不使用训练过的模型,使用随机初始化的模型
        if model_path == '' or model_path == None:
            print('Use a random model')
            self.if_load = True
        
        
    def Extractive(self, contexts, batch_size = 1, language = 'zh', len_pred = 2):
        """
        运行抽取式摘要过程。

        参数:
        - self: 类实例的引用。
        - contexts (list of str): 需要进行摘要提取的文本列表。
        - batch_size (int, optional): 在模型推理过程中一次处理的文本数量，默认为1。
        - language (str, optional): 输入文本的语言，默认为'zh'（中文）。
        - len_pred (int, optional): 预期生成摘要的句子数量，默认为2。

        返回:
        - condidate_list (list of str): 模型生成的摘要列表。

        过程描述:
        - 首先检查模型是否已加载，如果尚未加载，则加载模型并设置为评估模式。
        - 使用生成函数生成摘要，该函数接受模型、文本、设备等参数。
        - 返回生成的摘要列表。

        """
        if self.if_load == False:
            checkpoint = torch.load(self.model_path, map_location=lambda storage, loc: storage)
            self.model.load_cp(checkpoint)
            self.model.eval()
            self.if_load = True
            print(f'load model success. load from{self.model_path}')
        condidate_list = generate(self.model, contexts, self.device, batch_size=batch_size, len_pred=len_pred, language=language)
        
        return condidate_list
    

        


from setuptools import setup, find_packages

setup(
    name='CINOSUM',
    version='1.0.4',
    description='Chinese Minority Extractive Multi-language summarization project',
    author='HaoYu Luo',
    author_email='506685820@qq.com',
    packages=find_packages(),
    install_requires=[
        'torch',
        'numpy',
        'tqdm',
        'transformers==4.33.1',
    ],
    
    
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.8',
)


""" Optimizers class """
import torch
import torch.optim as optim
from torch.nn.utils import clip_grad_norm_


# from onmt.utils import use_gpu


def use_gpu(opt):
    """
    Creates a boolean if gpu used
    """
    return (hasattr(opt, 'gpu_ranks') and len(opt.gpu_ranks) > 0) or \
           (hasattr(opt, 'gpu') and opt.gpu > -1)

def build_optim(model, opt, checkpoint):
    """ Build optimizer """
    saved_optimizer_state_dict = None

    if opt.train_from:
        optim = checkpoint['optim']
        # We need to save a copy of optim.optimizer.state_dict() for setting
        # the, optimizer state later on in Stage 2 in this method, since
        # the method optim.set_parameters(model.parameters()) will overwrite
        # optim.optimizer, and with ith the values stored in
        # optim.optimizer.state_dict()
        saved_optimizer_state_dict = optim.optimizer.state_dict()
    else:
        optim = Optimizer(
            opt.optim, opt.learning_rate, opt.max_grad_norm,
            lr_decay=opt.learning_rate_decay,
            start_decay_steps=opt.start_decay_steps,
            decay_steps=opt.decay_steps,
            beta1=opt.adam_beta1,
            beta2=opt.adam_beta2,
            adagrad_accum=opt.adagrad_accumulator_init,
            decay_method=opt.decay_method,
            warmup_steps=opt.warmup_steps)

    # Stage 1:
    # Essentially optim.set_parameters (re-)creates and optimizer using
    # model.paramters() as parameters that will be stored in the
    # optim.optimizer.param_groups field of the torch optimizer class.
    # Importantly, this method does not yet load the optimizer state, as
    # essentially it builds a new optimizer with empty optimizer state and
    # parameters from the model.
    optim.set_parameters(model.named_parameters())

    if opt.train_from:
        # Stage 2: In this stage, which is only performed when loading an
        # optimizer from a checkpoint, we load the saved_optimizer_state_dict
        # into the re-created optimizer, to set the optim.optimizer.state
        # field, which was previously empty. For this, we use the optimizer
        # state saved in the "saved_optimizer_state_dict" variable for
        # this purpose.
        # See also: https://github.com/pytorch/pytorch/issues/2830
        optim.optimizer.load_state_dict(saved_optimizer_state_dict)
        # Convert back the state values to cuda type if applicable
        if use_gpu(opt):
            for state in optim.optimizer.state.values():
                for k, v in state.items():
                    if torch.is_tensor(v):
                        state[k] = v.cuda()

        # We want to make sure that indeed we have a non-empty optimizer state
        # when we loaded an existing model. This should be at least the case
        # for Adam, which saves "exp_avg" and "exp_avg_sq" state
        # (Exponential moving average of gradient and squared gradient values)
        if (optim.method == 'adam') and (len(optim.optimizer.state) < 1):
            raise RuntimeError(
                "Error: loaded Adam optimizer from existing model" +
                " but optimizer state is empty")

    return optim


class MultipleOptimizer(object):
    """ Implement multiple optimizers needed for sparse adam """

    def __init__(self, op):
        """ ? """
        self.optimizers = op

    def zero_grad(self):
        """ ? """
        for op in self.optimizers:
            op.zero_grad()

    def step(self):
        """ ? """
        for op in self.optimizers:
            op.step()

    @property
    def state(self):
        """ ? """
        return {k: v for op in self.optimizers for k, v in op.state.items()}

    def state_dict(self):
        """ ? """
        return [op.state_dict() for op in self.optimizers]

    def load_state_dict(self, state_dicts):
        """ ? """
        assert len(state_dicts) == len(self.optimizers)
        for i in range(len(state_dicts)):
            self.optimizers[i].load_state_dict(state_dicts[i])


class Optimizer(object):
    """
    Controller class for optimization. Mostly a thin
    wrapper for `optim`, but also useful for implementing
    rate scheduling beyond what is currently available.
    Also implements necessary methods for training RNNs such
    as grad manipulations.

    Args:
      method (:obj:`str`): one of [sgd, adagrad, adadelta, adam]
      lr (float): learning rate
      lr_decay (float, optional): learning rate decay multiplier
      start_decay_steps (int, optional): step to start learning rate decay
      beta1, beta2 (float, optional): parameters for adam
      adagrad_accum (float, optional): initialization parameter for adagrad
      decay_method (str, option): custom decay options
      warmup_steps (int, option): parameter for `noam` decay

    We use the default parameters for Adam that are suggested by
    the original paper https://arxiv.org/pdf/1412.6980.pdf
    These values are also used by other established implementations,
    e.g. https://www.tensorflow.org/api_docs/python/tf/train/AdamOptimizer
    https://keras.io/optimizers/
    Recently there are slightly different values used in the paper
    "Attention is all you need"
    https://arxiv.org/pdf/1706.03762.pdf, particularly the value beta2=0.98
    was used there however, beta2=0.999 is still arguably the more
    established value, so we use that here as well
    """

    def __init__(self, method, learning_rate, max_grad_norm,
                 lr_decay=1, start_decay_steps=None, decay_steps=None,
                 beta1=0.9, beta2=0.999,
                 adagrad_accum=0.0,
                 decay_method=None,
                 warmup_steps=4000
                 ):
        self.last_ppl = None
        self.learning_rate = learning_rate
        self.original_lr = learning_rate
        self.max_grad_norm = max_grad_norm
        self.method = method
        self.lr_decay = lr_decay
        self.start_decay_steps = start_decay_steps
        self.decay_steps = decay_steps
        self.start_decay = False
        self._step = 0
        self.betas = [beta1, beta2]
        self.adagrad_accum = adagrad_accum
        self.decay_method = decay_method
        self.warmup_steps = warmup_steps

    def set_parameters(self, params):
        """ ? """
        self.params = []
        self.sparse_params = []
        for k, p in params:
            if p.requires_grad:
                if self.method != 'sparseadam' or "embed" not in k:
                    self.params.append(p)
                else:
                    self.sparse_params.append(p)
        if self.method == 'sgd':
            self.optimizer = optim.SGD(self.params, lr=self.learning_rate)
        elif self.method == 'adagrad':
            self.optimizer = optim.Adagrad(self.params, lr=self.learning_rate)
            for group in self.optimizer.param_groups:
                for p in group['params']:
                    self.optimizer.state[p]['sum'] = self.optimizer\
                        .state[p]['sum'].fill_(self.adagrad_accum)
        elif self.method == 'adadelta':
            self.optimizer = optim.Adadelta(self.params, lr=self.learning_rate)
        elif self.method == 'adam':
            self.optimizer = optim.Adam(self.params, lr=self.learning_rate,
                                        betas=self.betas, eps=1e-9)
        elif self.method == 'sparseadam':
            self.optimizer = MultipleOptimizer(
                [optim.Adam(self.params, lr=self.learning_rate,
                            betas=self.betas, eps=1e-8),
                 optim.SparseAdam(self.sparse_params, lr=self.learning_rate,
                                  betas=self.betas, eps=1e-8)])
        else:
            raise RuntimeError("Invalid optim method: " + self.method)

    def _set_rate(self, learning_rate):
        self.learning_rate = learning_rate
        if self.method != 'sparseadam':
            self.optimizer.param_groups[0]['lr'] = self.learning_rate
        else:
            for op in self.optimizer.optimizers:
                op.param_groups[0]['lr'] = self.learning_rate

    def step(self):
        """Update the model parameters based on current gradients.

        Optionally, will employ gradient modification or update learning
        rate.
        """
        self._step += 1

        # Decay method used in tensor2tensor.
        if self.decay_method == "noam":
            self._set_rate(
                self.original_lr *

                 min(self._step ** (-0.5),
                     self._step * self.warmup_steps**(-1.5)))

            # self._set_rate(self.original_lr *self.model_size ** (-0.5) *min(1.0, self._step / self.warmup_steps)*max(self._step, self.warmup_steps)**(-0.5))
        # Decay based on start_decay_steps every decay_steps
        else:
            if ((self.start_decay_steps is not None) and (
                     self._step >= self.start_decay_steps)):
                self.start_decay = True
            if self.start_decay:
                if ((self._step - self.start_decay_steps)
                   % self.decay_steps == 0):
                    self.learning_rate = self.learning_rate * self.lr_decay

        if self.method != 'sparseadam':
            self.optimizer.param_groups[0]['lr'] = self.learning_rate

        if self.max_grad_norm:
            clip_grad_norm_(self.params, self.max_grad_norm)
        self.optimizer.step()


